import numpy as np
import numpy.testing as npt
import pytest
import tifffile
from lxml import etree  # nosec

from PartSegImage import ChannelInfo
from PartSegImage.image import Image
from PartSegImage.image_reader import TiffImageReader
from PartSegImage.image_writer import IMAGEJImageWriter, ImageWriter


@pytest.fixture(scope="module")
def ome_xml(bundle_test_dir):
    return etree.XMLSchema(file=str(bundle_test_dir / "ome.xsd.xml"))


def test_scaling(tmp_path):
    image = Image(np.zeros((10, 50, 50), dtype=np.uint8), spacing=(30, 0.1, 0.1), axes_order="ZYX")
    ImageWriter.save(image, tmp_path / "image.tif")
    read_image = TiffImageReader.read_image(tmp_path / "image.tif")
    assert np.all(np.isclose(image.spacing, read_image.spacing))


def test_save_mask(tmp_path):
    data = np.zeros((10, 40, 40), dtype=np.uint8)
    data[1:-1, 1:-1, 1:-1] = 1
    data[2:-3, 4:-4, 4:-4] = 2

    mask = np.array(data > 0).astype(np.uint8)

    image = Image(data, spacing=(0.4, 0.1, 0.1), mask=mask, axes_order="ZYX")
    ImageWriter.save_mask(image, tmp_path / "mask.tif")

    read_mask = TiffImageReader.read_image(tmp_path / "mask.tif")
    assert np.all(np.isclose(read_mask.spacing, image.spacing))
    image.set_mask(None)
    ImageWriter.save_mask(image, tmp_path / "mask2.tif")
    assert not (tmp_path / "mask2.tif").exists()


@pytest.mark.parametrize("z_size", [1, 10])
def test_ome_save(tmp_path, bundle_test_dir, ome_xml, z_size):
    data = np.zeros((z_size, 20, 20, 2), dtype=np.uint8)
    image = Image(
        data,
        spacing=(27 * 10**-6, 6 * 10**-6, 6 * 10**-6),
        axes_order="ZYXC",
        channel_info=[ChannelInfo(name="a"), ChannelInfo(name="b")],
        shift=(10, 9, 8),
        name="Test",
    )
    ImageWriter.save(image, tmp_path / "test.tif")

    with tifffile.TiffFile(tmp_path / "test.tif") as tiff:
        assert tiff.is_ome
        assert isinstance(tiff.ome_metadata, str)
        meta_data = tifffile.xml2dict(tiff.ome_metadata)["OME"]["Image"]
        assert "PhysicalSizeX" in meta_data["Pixels"]
        assert meta_data["Pixels"]["PhysicalSizeX"] == 6
        assert "PhysicalSizeXUnit" in meta_data["Pixels"]
        assert meta_data["Pixels"]["PhysicalSizeXUnit"] == "µm"
        assert len(meta_data["Pixels"]["Channel"]) == 2
        assert meta_data["Pixels"]["Channel"][0]["Name"] == "a"
        assert meta_data["Pixels"]["Channel"][1]["Name"] == "b"
        assert meta_data["Name"] == "Test"
        xml_file = etree.fromstring(tiff.ome_metadata.encode("utf8"))  # nosec  # noqa: S320
        ome_xml.assert_(xml_file)  # noqa: PT009
    read_image = TiffImageReader.read_image(tmp_path / "test.tif")
    assert np.allclose(read_image.spacing, image.spacing)
    assert np.allclose(read_image.shift, image.shift)
    assert read_image.channel_names == ["a", "b"]
    assert read_image.name == "Test"


def test_scaling_imagej(tmp_path):
    image = Image(np.zeros((10, 50, 50), dtype=np.uint8), spacing=(30, 0.1, 0.1), axes_order="ZYX")
    IMAGEJImageWriter.save(image, tmp_path / "image.tif")
    read_image = TiffImageReader.read_image(tmp_path / "image.tif")
    assert np.all(np.isclose(image.spacing, read_image.spacing))


def test_scaling_imagej_fail(tmp_path):
    image = Image(np.zeros((10, 50, 50), dtype=np.float64), spacing=(30, 0.1, 0.1), axes_order="ZYX")
    with pytest.raises(ValueError, match="Data type float64"):
        IMAGEJImageWriter.save(image, tmp_path / "image.tif")


def test_imagej_write_all_metadata(tmp_path, data_test_dir):
    image = TiffImageReader.read_image(data_test_dir / "stack1_components" / "stack1_component1.tif")
    assert "coloring: " in str(image)
    assert "coloring: None" not in str(image)
    IMAGEJImageWriter.save(image, tmp_path / "image.tif")

    image2 = TiffImageReader.read_image(tmp_path / "image.tif")

    npt.assert_array_equal(image2.default_coloring, image.get_imagej_colors())


def test_imagej_save_color(tmp_path):
    data = np.zeros((4, 20, 20), dtype=np.uint8)
    data[:, 2:-2, 2:-2] = 20
    img = Image(
        data,
        spacing=(0.4, 0.1, 0.1),
        axes_order="CYX",
        channel_info=[
            ChannelInfo(name="ch1", color_map="blue", contrast_limits=(0, 20)),
            ChannelInfo(name="ch2", color_map="#FFAA00", contrast_limits=(0, 30)),
            ChannelInfo(name="ch3", color_map="#FB1", contrast_limits=(0, 25)),
            ChannelInfo(name="ch4", color_map=(0, 180, 0), contrast_limits=(0, 22)),
        ],
    )
    assert img.get_colors()[:3] == ["blue", "#FFAA00", "#FB1"]
    assert tuple(img.get_colors()[3]) == (0, 180, 0)
    IMAGEJImageWriter.save(img, tmp_path / "image.tif")
    image2 = TiffImageReader.read_image(tmp_path / "image.tif")
    assert image2.channel_names == ["ch1", "ch2", "ch3", "ch4"]
    assert image2.ranges == [(0, 20), (0, 30), (0, 25), (0, 22)]
    assert tuple(image2.default_coloring[0][:, -1]) == (0, 0, 255)
    assert tuple(image2.default_coloring[1][:, -1]) == (255, 170, 0)
    assert tuple(image2.default_coloring[2][:, -1]) == (255, 187, 17)
    assert tuple(image2.default_coloring[3][:, -1]) == (0, 180, 0)


def test_save_mask_imagej(tmp_path):
    data = np.zeros((10, 40, 40), dtype=np.uint8)
    data[1:-1, 1:-1, 1:-1] = 1
    data[2:-3, 4:-4, 4:-4] = 2

    mask = np.array(data > 0).astype(np.uint8)

    image = Image(data, spacing=(0.4, 0.1, 0.1), mask=mask, axes_order="ZYX")
    IMAGEJImageWriter.save_mask(image, tmp_path / "mask.tif")

    read_mask = TiffImageReader.read_image(tmp_path / "mask.tif")
    assert np.all(np.isclose(read_mask.spacing, image.spacing))



"""
This module contains PartSeg function used for calculate in batch processing

Calculation hierarchy:

.. graphviz::

   digraph calc {
      rankdir="LR";
      "CalculationManager"[shape=rectangle style=filled];
      "BatchManager"[shape=rectangle];
      "BatchWorker"[shape=rectangle];
      "CalculationManager" -> "BatchManager" -> "BatchWorker"[arrowhead="crow"];

      "CalculationManager" -> "DataWriter"[arrowhead="inv"];
      "DataWriter"[shape=rectangle];
      "FileData"[shape=rectangle];
      "SheetData"[shape=rectangle];
      "DataWriter" -> "FileData" -> "SheetData"[arrowhead="crow"];
   }

"""

from __future__ import annotations

import contextlib
import json
import logging
import math
import os
import threading
import traceback
from collections import OrderedDict
from enum import Enum
from os import path
from queue import Queue
from traceback import StackSummary
from typing import TYPE_CHECKING, Any, NamedTuple, Union

import numpy as np
import pandas as pd
import SimpleITK
import tifffile

from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis.algorithm_description import AnalysisAlgorithmSelection
from PartSegCore.analysis.batch_processing.parallel_backend import BatchManager, SubprocessOrder
from PartSegCore.analysis.calculation_plan import (
    BaseCalculation,
    Calculation,
    CalculationPlan,
    CalculationTree,
    FileCalculation,
    MaskCreate,
    MaskIntersection,
    MaskMapper,
    MaskSum,
    MaskUse,
    MeasurementCalculate,
    Operations,
    RootType,
    Save,
    get_save_path,
)
from PartSegCore.analysis.io_utils import ProjectTuple
from PartSegCore.analysis.load_functions import LoadImageForBatch, LoadMaskSegmentation, LoadProject
from PartSegCore.analysis.measurement_base import has_mask_components, has_roi_components
from PartSegCore.analysis.save_functions import save_dict
from PartSegCore.json_hooks import PartSegEncoder
from PartSegCore.mask_create import calculate_mask
from PartSegCore.project_info import AdditionalLayerDescription, HistoryElement
from PartSegCore.roi_info import ROIInfo
from PartSegCore.segmentation.algorithm_base import ROIExtractionAlgorithm, report_empty_fun
from PartSegCore.utils import iterate_names
from PartSegImage import Image, TiffImageReader

if TYPE_CHECKING:
    import uuid

    import xlsxwriter

    from PartSegCore.analysis.measurement_calculation import MeasurementResult
    from PartSegCore.segmentation import RestartableAlgorithm

# https://support.microsoft.com/en-us/office/excel-specifications-and-limits-1672b34d-7043-467e-8e27-269d656771c3#ID0EDBD=Newer_versions
# page with excel limits
MAX_CHAR_IN_EXCEL_CELL = 30_000  # real limit is 32_767 but it is better to have some margin
MAX_ROWS_IN_EXCEL_CELL = 50  # real limit is 253 but 50 provides better readability


class ResponseData(NamedTuple):
    path_to_file: str
    values: list[MeasurementResult]


CalculationResultList = list[ResponseData]
ErrorInfo = tuple[Exception, Union[StackSummary, tuple[dict, StackSummary]]]
WrappedResult = tuple[int, list[Union[ErrorInfo, ResponseData]]]


def get_data_loader(
    root_type: RootType, file_path: str
) -> tuple[type[LoadMaskSegmentation | LoadProject | LoadImageForBatch], bool]:
    """
    Get data loader for given root type. Return indicator if file extension match to loader.

    :param RootType root_type: type of loader
    :param str file_path: path to file
    :return: Loader and indicator if file extension match to loader
    """

    ext = path.splitext(file_path)[1].lower()
    if root_type == RootType.Mask_project:
        return LoadMaskSegmentation, ext in LoadMaskSegmentation.get_extensions()
    if root_type == RootType.Project:
        return LoadProject, ext in LoadProject.get_extensions()
    return LoadImageForBatch, ext in LoadImageForBatch.get_extensions()


def prepare_error_data(exception: Exception) -> ErrorInfo:
    try:
        from sentry_sdk.serializer import serialize
        from sentry_sdk.utils import event_from_exception

        event = event_from_exception(exception)[0]
        event = serialize(event)
        return exception, (event, traceback.extract_tb(exception.__traceback__))
    except ImportError:  # pragma: no cover
        return exception, traceback.extract_tb(exception.__traceback__)


def do_calculation(file_info: tuple[int, str], calculation: BaseCalculation) -> WrappedResult:
    """
    Main function which will be used for run calculation.
    It create :py:class:`.CalculationProcess` and call it method
    :py:meth:`.CalculationProcess.do_calculation`

    :param file_info: index and path to file which should be processed
    :param calculation: calculation description
    """
    with contextlib.suppress(AttributeError):
        SimpleITK.ProcessObject_SetGlobalDefaultNumberOfThreads(1)
    calc = CalculationProcess()
    index, file_path = file_info
    try:
        return index, calc.do_calculation(FileCalculation(file_path, calculation))
    except Exception as e:  # pylint: disable=broad-except
        return index, [prepare_error_data(e)]


class CalculationProcess:
    """
    Main class to calculate PartSeg calculation plan.
    To support other operations overwrite :py:meth:`recursive_calculation`
    call super function to support already defined operations.
    """

    def __init__(self):
        self.reused_mask = set()
        self.mask_dict = {}
        self.calculation = None
        self.measurement: list[MeasurementResult] = []
        self.image: Image | None = None
        self.roi_info: ROIInfo | None = None
        self.additional_layers: dict[str, AdditionalLayerDescription] = {}
        self.mask: np.ndarray | None = None
        self.history: list[HistoryElement] = []
        self.algorithm_parameters: dict = {}
        self.results: CalculationResultList = []

    def _reset_image_cache(self):
        self.image = None
        self.roi_info = None
        self.additional_layers = {}
        self.mask = None
        self.history = []
        self.algorithm_parameters = {}
        self.measurement = []
        self.reused_mask = set()

    @staticmethod
    def load_data(operation, calculation: FileCalculation) -> ProjectTuple | list[ProjectTuple]:
        metadata = {"default_spacing": calculation.voxel_size}

        loader, ext_match = get_data_loader(operation, calculation.file_path)

        try:
            return loader.load([calculation.file_path], metadata=metadata)
        except Exception as e:  # pragma: no cover
            if ext_match:
                raise e
            raise ValueError(f"File {calculation.file_path} do not match to {operation}") from e

    def do_calculation(self, calculation: FileCalculation) -> CalculationResultList:
        """
        Main function for calculation process

        :param calculation: calculation to do.
        :return:
        """
        self.calculation = calculation
        self.reused_mask = calculation.calculation_plan.get_reused_mask()
        self.mask_dict = {}
        self.measurement = []
        self.results = []
        operation = calculation.calculation_plan.execution_tree.operation
        projects = self.load_data(operation, calculation)

        if isinstance(projects, ProjectTuple):
            projects = [projects]
        for project in projects:
            try:
                self.image = project.image
                if calculation.overwrite_voxel_size:
                    self.image.set_spacing(calculation.voxel_size)
                if operation == RootType.Mask_project:
                    self.mask = project.mask
                if operation == RootType.Project:
                    self.mask = project.mask
                    # FIXME when load annotation from project is done
                    self.roi_info = project.roi_info
                    self.additional_layers = project.additional_layers
                    self.history = project.history
                    self.algorithm_parameters = project.algorithm_parameters

                self.iterate_over(calculation.calculation_plan.execution_tree)
                for el in self.measurement:
                    el.set_filename(path.relpath(project.image.file_path, calculation.base_prefix))
                self.results.append(
                    ResponseData(path.relpath(project.image.file_path, calculation.base_prefix), self.measurement)
                )
            except Exception as e:  # pylint: disable=broad-except
                self.results.append(prepare_error_data(e))
            self._reset_image_cache()
        return self.results

    def iterate_over(self, node: CalculationTree | list[CalculationTree]):
        """
        Execute calculation on node children or list oof nodes

        :type node: CalculationTree
        :param node:
        :return:
        """
        if isinstance(node, CalculationTree):
            node = node.children
        for el in node:
            self.recursive_calculation(el)

    def step_load_mask(self, operation: MaskMapper, children: list[CalculationTree]):
        """
        Load mask using mask mapper (mask can be defined with suffix, substitution, or file with mapping saved,
        then iterate over ``children`` nodes.

        :param MaskMapper operation: operation to perform
        :param List[CalculationTree] children: list of nodes to iterate over with applied mask
        """
        mask_path = operation.get_mask_path(self.calculation.file_path)
        if not mask_path:  # pragma: no cover
            raise ValueError("Empty path to mask.")
        if not os.path.exists(mask_path):
            raise OSError(f"Mask file {mask_path} does not exists")
        with tifffile.TiffFile(mask_path) as mask_file:
            mask = mask_file.asarray()
            mask = TiffImageReader.update_array_shape(mask, mask_file.series[0].axes)
            if "C" in TiffImageReader.image_class.axis_order:
                pos: list[slice | int] = [slice(None) for _ in range(mask.ndim)]
                pos[TiffImageReader.image_class.axis_order.index("C")] = 0
                mask = mask[tuple(pos)]

        mask = (mask > 0).astype(np.uint8)
        try:
            mask = self.image.fit_array_to_image(mask)[0]
            # TODO fix this time bug fix
        except ValueError as e:  # pragma: no cover
            raise ValueError("Mask do not fit to given image") from e
        old_mask = self.mask
        self.mask = mask
        self.iterate_over(children)
        self.mask = old_mask

    def step_segmentation(self, operation: ROIExtractionProfile, children: list[CalculationTree]):
        """
        Perform segmentation and iterate over ``children`` nodes

        :param ROIExtractionProfile operation: Specification of segmentation operation
        :param List[CalculationTree] children: list of nodes to iterate over after perform segmentation
        """
        segmentation_class = AnalysisAlgorithmSelection.get(operation.algorithm)
        if segmentation_class is None:  # pragma: no cover
            raise ValueError(f"Segmentation class {operation.algorithm} do not found")
        segmentation_algorithm: RestartableAlgorithm = segmentation_class()
        segmentation_algorithm.set_image(self.image)
        segmentation_algorithm.set_mask(self.mask)
        if segmentation_algorithm.__new_style__:
            segmentation_algorithm.set_parameters(operation.values)
        else:
            segmentation_algorithm.set_parameters(**operation.values)
        result = segmentation_algorithm.calculation_run(report_empty_fun)
        backup_data = self.roi_info, self.additional_layers, self.algorithm_parameters
        self.roi_info = ROIInfo(result.roi, result.roi_annotation, result.alternative_representation)
        self.additional_layers = result.additional_layers
        self.algorithm_parameters = {"algorithm_name": operation.algorithm, "values": operation.values}
        self.iterate_over(children)
        self.roi_info, self.additional_layers, self.algorithm_parameters = backup_data

    def step_mask_use(self, operation: MaskUse, children: list[CalculationTree]):
        """
        use already defined mask and iterate over ``children`` nodes

        :param MaskUse operation:
        :param List[CalculationTree] children: list of nodes to iterate over after perform segmentation
        """
        old_mask = self.mask
        mask = self.mask_dict[operation.name]
        self.mask = mask
        self.iterate_over(children)
        self.mask = old_mask

    def step_mask_operation(self, operation: MaskSum | MaskIntersection, children: list[CalculationTree]):
        """
        Generate new mask by sum or intersection of existing and iterate over ``children`` nodes

        :param operation: mask operation to perform
        :type operation: Union[MaskSum, MaskIntersection]
        :param List[CalculationTree] children: list of nodes to iterate over after perform segmentation
        """
        old_mask = self.mask
        mask1 = self.mask_dict[operation.mask1]
        mask2 = self.mask_dict[operation.mask2]
        if isinstance(operation, MaskSum):
            mask = np.logical_or(mask1, mask2).astype(np.uint8)
        else:
            mask = np.logical_and(mask1, mask2).astype(np.uint8)
        self.mask = mask
        self.iterate_over(children)
        self.mask = old_mask

    def step_save(self, operation: Save):
        """
        Perform save operation selected in plan.

        :param Save operation: save definition
        """
        save_class = save_dict[operation.algorithm]
        project_tuple = ProjectTuple(
            file_path="",
            image=self.image,
            roi_info=self.roi_info,
            additional_layers=self.additional_layers,
            mask=self.mask,
            history=self.history,
            algorithm_parameters=self.algorithm_parameters,
        )
        save_path = get_save_path(operation, self.calculation)
        save_dir = os.path.dirname(save_path)
        os.makedirs(save_dir, exist_ok=True)
        save_class.save(save_path, project_tuple, operation.values)

    def step_mask_create(self, operation: MaskCreate, children: list[CalculationTree]):
        """
        Create mask from current segmentation state using definition

        :param MaskCreate operation: mask create description.
        :param List[CalculationTree] children: list of nodes to iterate over after perform segmentation
        """
        mask = calculate_mask(
            mask_description=operation.mask_property,
            roi=self.roi_info.roi,
            old_mask=self.mask,
            spacing=self.image.spacing,
            time_axis=self.image.time_pos,
        )
        if operation.name in self.reused_mask:
            self.mask_dict[operation.name] = mask
        history_element = HistoryElement.create(
            self.roi_info,
            self.mask,
            self.algorithm_parameters,
            operation.mask_property,
        )
        backup = self.mask, self.history
        self.mask = mask
        self.history.append(history_element)
        self.iterate_over(children)
        self.mask, self.history = backup

    def step_measurement(self, operation: MeasurementCalculate):
        """
        Calculate measurement defined in current operation.

        :param MeasurementCalculate operation: definition of measurement to calculate
        """
        channel = operation.channel
        if channel == -1:
            segmentation_class: type[ROIExtractionAlgorithm] = AnalysisAlgorithmSelection.get(
                self.algorithm_parameters["algorithm_name"]
            )
            if segmentation_class is None:  # pragma: no cover
                raise ValueError(f"Segmentation class {self.algorithm_parameters['algorithm_name']} do not found")
            if segmentation_class.__new_style__:
                channel = getattr(self.algorithm_parameters["values"], segmentation_class.get_channel_parameter_name())
            else:
                channel = self.algorithm_parameters["values"][segmentation_class.get_channel_parameter_name()]

        # FIXME use additional information
        old_mask = self.image.mask
        self.image.set_mask(self.mask)
        measurement = operation.measurement_profile.calculate(
            self.image,
            channel,
            self.roi_info,
            operation.units,
        )
        self.measurement.append(measurement)
        self.image.set_mask(old_mask)

    def recursive_calculation(self, node: CalculationTree):
        """
        Identify node type and then call proper ``step_*`` function

        :param CalculationTree node: Node to be proceed
        """
        if isinstance(node.operation, MaskMapper):
            self.step_load_mask(node.operation, node.children)
        elif isinstance(node.operation, ROIExtractionProfile):
            self.step_segmentation(node.operation, node.children)
        elif isinstance(node.operation, MaskUse):
            self.step_mask_use(node.operation, node.children)
        elif isinstance(node.operation, (MaskSum, MaskIntersection)):
            self.step_mask_operation(node.operation, node.children)
        elif isinstance(node.operation, Save):
            self.step_save(node.operation)
        elif isinstance(node.operation, MaskCreate):
            self.step_mask_create(node.operation, node.children)
        elif isinstance(node.operation, Operations):  # pragma: no cover
            # backward compatibility
            self.iterate_over(node)
        elif isinstance(node.operation, MeasurementCalculate):
            self.step_measurement(node.operation)
        else:  # pragma: no cover
            raise ValueError(f"Unknown operation {type(node.operation)} {node.operation}")


class BatchResultDescription(NamedTuple):
    """
    Tuple to handle information about part of calculation result.
    """

    errors: list[tuple[str, ErrorInfo]]  #: list of errors occurred during calculation
    global_counter: int  #: total number of calculated steps
    jobs_status: dict[uuid.UUID, int]  #: for each job information about progress


class CalculationManager:
    """
    This class manage batch processing in PartSeg.

    """

    def __init__(self):
        self.batch_manager = BatchManager()
        self.calculation_queue = Queue()
        self.calculation_dict: dict[uuid.UUID, Calculation] = OrderedDict()
        self.calculation_sizes = []
        self.calculation_size = 0
        self.calculation_done = 0
        self.counter_dict = OrderedDict()
        self.errors_list = []
        self.writer = DataWriter()

    def is_valid_sheet_name(self, excel_path: str, sheet_name: str) -> bool:
        """
        Check if sheet name can be used

        :param str excel_path: path which allow identify excel file
        :param str sheet_name: name of excel sheet
        :return:
        :rtype: bool
        """
        return self.writer.is_empty_sheet(excel_path, sheet_name)

    def remove_calculation(self, calculation: Calculation):
        size = len(calculation.file_list)
        self.calculation_size -= size
        self.writer.remove_data_part(calculation)

    def cancel_calculation(self, calculation: Calculation):
        self.batch_manager.cancel_work(calculation)

    def add_calculation(self, calculation: Calculation):
        """
        :param calculation: Calculation
        """
        self.calculation_dict[calculation.uuid] = calculation
        self.counter_dict[calculation.uuid] = 0
        size = len(calculation.file_list)
        self.calculation_sizes.append(size)
        self.calculation_size += size
        self.batch_manager.add_work(
            list(enumerate(calculation.file_list)), calculation.get_base_calculation(), do_calculation
        )
        self.writer.add_data_part(calculation)

    @property
    def has_work(self) -> bool:
        """
        Is still some calculation or data writing in progress
        """
        return self.batch_manager.has_work or not self.writer.writing_finished()

    def kill_jobs(self):
        self.batch_manager.kill_jobs()

    def set_number_of_workers(self, val: int):
        """
        Set number of workers to perform calculation.

        :param int val: number of workers.
        """
        logging.debug("Number off process %s", val)
        self.batch_manager.set_number_of_process(val)

    def get_results(self) -> BatchResultDescription:
        """
        Consume results from :py:class:`BatchWorker` and transfer it to :py:class:`DataWriter`

        :return: information about calculation status
        :rtype: BatchResultDescription
        """
        responses: list[tuple[uuid.UUID, WrappedResult]] = self.batch_manager.get_result()
        new_errors: list[tuple[str, ErrorInfo]] = []
        for uuid_id, (ind, result_list) in responses:
            if uuid_id == "-1":  # pragma: no cover
                self.errors_list.append((f"Unknown file {ind}", result_list))
                new_errors.append((f"Unknown file {ind}", result_list))
                continue

            self.calculation_done += 1
            self.counter_dict[uuid_id] += 1
            calculation = self.calculation_dict[uuid_id]
            for el in result_list:
                if isinstance(el, ResponseData):
                    errors = self.writer.add_result(el, calculation, ind=ind)
                    new_errors.extend((el.path_to_file, err) for err in errors)
                elif el != SubprocessOrder.cancel_job:
                    file_info = calculation.file_list[ind] if ind != -1 else "unknown file"
                    self.writer.add_calculation_error(calculation, file_info, el[0])
                    self.errors_list.append((file_info, el))
                    new_errors.append((file_info, el))

                if self.counter_dict[uuid_id] == len(calculation.file_list):
                    errors = self.writer.calculation_finished(calculation)
                    new_errors.extend(("", err) for err in errors)
        return BatchResultDescription(new_errors, self.calculation_done, self.counter_dict.copy())


class FileType(Enum):
    excel_xlsx_file = 1
    excel_xls_file = 2
    text_file = 3


class SheetData:
    """
    Store single sheet information
    """

    def __init__(self, name: str, columns: list[tuple[str, str]], raw=False):
        if len(columns) != len(set(columns)):
            raise ValueError(f"Columns should be unique: {columns}")
        self.name = name
        if raw:
            self.columns = pd.MultiIndex.from_tuples(columns)
        else:
            self.columns = pd.MultiIndex.from_tuples([("name", "units"), *columns])
        self.data_frame = pd.DataFrame([], columns=self.columns)
        self.row_list: list[Any] = []

    def add_data(self, data, ind):
        if len(data) != len(self.columns):
            raise ValueError(
                f"Wrong number of columns in data ({len(data)} instead of "
                f"{len(self.columns)} {data} for columns {self.columns.values}"
            )
        if ind is None:
            ind = len(self.row_list)
        self.row_list.append((ind, data))

    def add_data_list(self, data, ind):
        if ind is None:
            ind = len(self.row_list)
        for x in data:
            self.add_data(x, ind)

    def get_data_to_write(self) -> tuple[str, pd.DataFrame]:
        """
        Get data for write

        :return: sheet name and data to write
        :rtype: Tuple[str, pd.DataFrame]
        """
        sorted_row = [x[1] for x in sorted(self.row_list)]
        df = pd.DataFrame(sorted_row, columns=self.columns)
        if self.data_frame.empty:
            self.data_frame = df.reset_index(drop=True)
        else:
            df2 = pd.concat((self.data_frame, df), axis=0)
            self.data_frame = df2.reset_index(drop=True)
        self.row_list = []
        return self.name, self.data_frame

    def __repr__(self):
        return f"SheetData(name={self.name}, columns{list(self.columns)[1:]}, wait_rows={len(self.row_list)})"


class FileData:
    """
    Handle information about single file.

    This class run separate thread for writing purpose.
    This need additional synchronisation. but not freeze

    :param BaseCalculation calculation: calculation information
    :param int write_threshold: every how many lines of data are written to disk
    :cvar component_str: separator for per component sheet information
    """

    component_str = "_comp_"  #: separator for per component sheet information

    def __init__(self, calculation: BaseCalculation, write_threshold: int = 40):
        """
        :param BaseCalculation calculation: calculation information
        :param int write_threshold: every how many lines of data are written to disk
        """
        self.file_path = calculation.measurement_file_path
        ext = path.splitext(calculation.measurement_file_path)[1]
        if ext == ".xlsx":
            self.file_type = FileType.excel_xlsx_file
        elif ext == ".xls":  # pragma: no cover
            self.file_type = FileType.excel_xls_file
        else:  # pragma: no cover
            self.file_type = FileType.text_file
        self.writing = False
        data = SheetData("calculation_info", [("Description", "str"), ("JSON", "str")], raw=True)
        data.add_data(
            [str(calculation.calculation_plan), json.dumps(calculation.calculation_plan, cls=PartSegEncoder)], 0
        )
        self.sheet_dict = {}
        self.calculation_info = {}
        self.sheet_set = {"Errors"}
        self.new_count = 0
        self.write_threshold = write_threshold
        self.wrote_queue = Queue()
        self.error_queue = Queue()
        self.write_thread = threading.Thread(target=self.wrote_data_to_file)
        self.write_thread.daemon = True
        self.write_thread.start()
        self._error_info = []
        self.add_data_part(calculation)

    def finished(self):
        """check if any data wait on write to disc"""
        return not self.writing and self.wrote_queue.empty()

    def good_sheet_name(self, name: str) -> tuple[bool, str]:
        """
        Check if sheet name can be used in current file.
        Return False if:

        * file is text file
        * contains :py:attr:`component_str` in name
        * name is already in use

        :param str name: sheet name
        :return: if can be used and error message
        """
        if self.file_type == FileType.text_file:
            return False, "Text file allow store only one sheet"
        if self.component_str in name:
            return False, f"Sequence '{FileData.component_str}' is reserved for auto generated sheets"
        if name in self.sheet_set:
            return False, f"Sheet name {name} already in use"
        return True, ""

    def remove_data_part(self, calculation: BaseCalculation):
        if calculation.uuid in self.sheet_dict:
            sheet_list = self.sheet_dict[calculation.uuid][1]
            for sheet in sheet_list:
                if sheet is None:
                    continue
                self.sheet_set.remove(sheet.name)
            self.sheet_set.remove(calculation.sheet_name)
            del self.sheet_dict[calculation.uuid]

        if calculation.uuid in self.calculation_info:
            del self.calculation_info[calculation.uuid]

    def add_data_part(self, calculation: BaseCalculation):
        """
        Add new calculation which result will be stored in handled file.

        :param BaseCalculation calculation: information about calculation
        :raises ValueError: when :py:attr:`measurement_file_path` is different to handled file
            or :py:attr:`sheet_name` name already is in use.
        """
        if calculation.measurement_file_path != self.file_path:
            raise ValueError(f"[FileData] different file path {calculation.measurement_file_path} vs {self.file_path}")
        if calculation.sheet_name in self.sheet_set:  # pragma: no cover
            raise ValueError(f"[FileData] sheet name {calculation.sheet_name} already in use")
        measurement = calculation.calculation_plan.get_measurements()
        component_information = [x.measurement_profile.get_component_info(x.units) for x in measurement]
        num = 1
        sheet_list = []
        header_list = []
        main_header = []
        for i, el in enumerate(component_information):
            local_header = []
            component_seg = has_roi_components(measurement[i].measurement_profile.get_component_and_area_info())
            component_mask = has_mask_components(measurement[i].measurement_profile.get_component_and_area_info())
            if component_seg:
                local_header.append(("Segmentation component", "num"))
            if component_mask:
                local_header.append(("Mask component", "num"))
            if any(x[1] for x in el):
                sheet_list.append(
                    f"{calculation.sheet_name}{FileData.component_str}{num} - "
                    f"{measurement[i].name_prefix + measurement[i].name}"
                )

                num += 1
            else:
                sheet_list.append(None)
            for name, comp in el:
                local_header.append(name)
                if not comp:
                    main_header.append(name)
            header_list.append(local_header)

        self.sheet_dict[calculation.uuid] = (
            SheetData(calculation.sheet_name, main_header),
            [SheetData(name, header_list[i]) if name is not None else None for i, name in enumerate(sheet_list)],
            component_information,
        )
        self.sheet_set.add(calculation.sheet_name)
        self.sheet_set.update(sheet_list)
        self.calculation_info[calculation.uuid] = calculation.calculation_plan

    def wrote_data(self, uuid_id: uuid.UUID, data: ResponseData, ind: int | None = None):
        """
        Add information to be stored in output file

        :param uuid.UUID uuid_id: calculation identifier
        :param ResponseData data: calculation result
        :param Optional[int] ind: element index
        """
        self.new_count += 1
        main_sheet, component_sheets, _component_information = self.sheet_dict[uuid_id]
        name = data.path_to_file
        data_list = [name]
        for el, comp_sheet in zip(data.values, component_sheets):
            data_list.extend(el.get_global_parameters()[1:])
            comp_list = el.get_separated()
            if comp_sheet is not None:
                comp_sheet.add_data_list(comp_list, ind)
        main_sheet.add_data(data_list, ind)
        if self.new_count >= self.write_threshold:
            self.dump_data()
            self.new_count = 0

    def wrote_errors(self, file_path, error_description):
        self.new_count += 1
        self._error_info.append((file_path, str(error_description)))

    def dump_data(self):
        """
        Fire writing data to disc
        """
        data = []
        for main_sheet, component_sheets, _ in self.sheet_dict.values():
            data.append(main_sheet.get_data_to_write())
            data.extend(sheet.get_data_to_write() for sheet in component_sheets if sheet is not None)

        self.wrote_queue.put((data, list(self.calculation_info.values()), self._error_info[:]))

    def wrote_data_to_file(self):
        """
        Main function to write data to hard drive.
        It is executed in separate thread.
        """
        while True:
            data = self.wrote_queue.get()
            if data == "finish":
                break
            self.writing = True
            try:
                if self.file_type == FileType.text_file:
                    base_path, ext = path.splitext(self.file_path)
                    for sheet_name, data_frame in data[0]:
                        data_frame.to_csv(f"{base_path}_{sheet_name}{ext}")
                    continue
                file_path = self.file_path
                i = 0
                while i < 100:
                    i += 1
                    try:
                        self.write_to_excel(file_path, data)
                        break
                    except OSError:
                        base, ext = path.splitext(self.file_path)
                        file_path = f"{base}({i}){ext}"
                if i == 100:  # pragma: no cover
                    raise PermissionError(f"Fail to write result excel {self.file_path}")
            except Exception as e:  # pragma: no cover   # pylint: disable=broad-except
                logging.error("[batch_backend] %s", e)
                self.error_queue.put(prepare_error_data(e))
            finally:
                self.writing = False

    @classmethod
    def write_to_excel(
        cls, file_path: str, data: tuple[list[tuple[str, pd.DataFrame]], list[CalculationPlan], list[tuple[str, str]]]
    ):
        with pd.ExcelWriter(file_path) as writer:  # pylint: disable=abstract-class-instantiated
            new_sheet_names = []
            ind = 0
            sheets, plans, errors = data
            for sheet_name, _ in sheets:
                if len(sheet_name) < 32:
                    new_sheet_names.append(sheet_name)
                else:
                    new_sheet_names.append(f"{sheet_name[:27]}_{ind}_")
                    ind += 1
            for sheet_name, (_, data_frame) in zip(new_sheet_names, sheets):
                data_frame.to_excel(writer, sheet_name=sheet_name)
                sheet = writer.book.sheetnames[sheet_name]
                sheet.set_column(1, 1, 10)
                for i, (text, _unit) in enumerate(data_frame.columns[1:], start=2):
                    sheet.set_column(i, i, len(text) + 1)

            for calculation_plan in plans:
                cls.write_calculation_plan(writer, calculation_plan)

            if errors:
                errors_data = pd.DataFrame(errors, columns=["File path", "error description"])
                errors_data.to_excel(writer, sheet_name="Errors")

    @staticmethod
    def write_calculation_plan(writer: pd.ExcelWriter, calculation_plan: CalculationPlan):
        book: xlsxwriter.Workbook = writer.book
        sheet_name = iterate_names(f"info {calculation_plan.name}"[:30], book.sheetnames, 30)
        if sheet_name is None:  # pragma: no cover
            raise ValueError(
                "Name collision in sheets with information about calculation "
                f"plan: {f'info {calculation_plan.name}'[:30]}"
            )

        sheet = book.add_worksheet(sheet_name)
        cell_format = book.add_format({"bold": True})
        sheet.write("A1", "Plan Description", cell_format)
        sheet.write("B1", "Plan JSON", cell_format)
        sheet.write("C1", "Plan JSON (readable)", cell_format)
        description = calculation_plan.pretty_print().split("\n")
        for i in range(math.ceil(len(description) / MAX_ROWS_IN_EXCEL_CELL)):
            to_write = description[i * MAX_ROWS_IN_EXCEL_CELL : (i + 1) * MAX_ROWS_IN_EXCEL_CELL]
            sheet.write(f"A{i+2}", "\n".join(to_write))
            sheet.set_row(i + 1, len(to_write) * 11 + 10)

        sheet.set_column(0, 0, max(map(len, description)))
        sheet.set_column(1, 1, 15)
        calculation_plan_str = json.dumps(calculation_plan, cls=PartSegEncoder)
        for i in range(math.ceil(len(calculation_plan_str) / MAX_CHAR_IN_EXCEL_CELL)):
            sheet.write(f"B{i+2}", calculation_plan_str[i * MAX_CHAR_IN_EXCEL_CELL : (i + 1) * MAX_CHAR_IN_EXCEL_CELL])

        calculation_plan_pretty = json.dumps(calculation_plan, cls=PartSegEncoder, indent=2).split("\n")
        for i in range(math.ceil(len(calculation_plan_pretty) / MAX_ROWS_IN_EXCEL_CELL)):
            to_write = calculation_plan_pretty[i * MAX_ROWS_IN_EXCEL_CELL : (i + 1) * MAX_ROWS_IN_EXCEL_CELL]
            sheet.write(f"C{i+2}", "\n".join(to_write))
            sheet.set_row(i + 1, len(to_write) * 11 + 10)

        sheet.set_column(2, 2, max(map(len, calculation_plan_pretty)))

    def get_errors(self) -> list[ErrorInfo]:
        """
        Get list of errors occurred in last write
        """
        res = []
        while not self.error_queue.empty():
            res.append(self.error_queue.get())
        return res

    def finish(self):
        self.wrote_queue.put("finish")

    def is_empty_sheet(self, sheet_name) -> bool:
        return self.good_sheet_name(sheet_name)[0]


class DataWriter:
    """
    Handle information
    """

    def __init__(self):
        self.file_dict: dict[str, FileData] = {}

    def is_empty_sheet(self, file_path: str, sheet_name: str) -> bool:
        """
        Check if given pair of `file_path` and `sheet_name` can be used.

        :param str file_path: path to file to store measurement result
        :param str sheet_name: Name of excel sheet in which data will be stored
        :return: If calling :py:meth:`FileData.add_data_part` finish without error.
        :rtype: bool
        """
        if FileData.component_str in sheet_name:
            return False
        if file_path not in self.file_dict:
            return True
        return self.file_dict[file_path].is_empty_sheet(sheet_name)

    def add_data_part(self, calculation: BaseCalculation):
        """
        Add information about calculation
        """
        if calculation.measurement_file_path in self.file_dict:
            self.file_dict[calculation.measurement_file_path].add_data_part(calculation)
        else:
            self.file_dict[calculation.measurement_file_path] = FileData(calculation)

    def remove_data_part(self, calculation: BaseCalculation):
        if calculation.measurement_file_path in self.file_dict:
            self.file_dict[calculation.measurement_file_path].remove_data_part(calculation)

    def add_result(self, data: ResponseData, calculation: BaseCalculation, ind: int | None = None) -> list[ErrorInfo]:
        """
        Add calculation result to file writer

        :raises ValueError: when calculation.measurement_file_path is not added with :py:meth:`.add_data_part`
        """
        if calculation.measurement_file_path not in self.file_dict:
            raise ValueError("Unknown measurement file")
        file_writer = self.file_dict[calculation.measurement_file_path]
        file_writer.wrote_data(calculation.uuid, data, ind)
        return file_writer.get_errors()

    def add_calculation_error(self, calculation: BaseCalculation, file_path: str, error):
        file_writer = self.file_dict[calculation.measurement_file_path]
        file_writer.wrote_errors(file_path, error)

    def writing_finished(self) -> bool:
        """check if all data are written to disc"""
        return all(x.finished() for x in self.file_dict.values())

    def finish(self):
        """close all files"""
        for file_data in self.file_dict.values():
            file_data.finish()

    def calculation_finished(self, calculation) -> list[ErrorInfo]:
        """
        Force write data for given calculation.

        :raises ValueError: when measurement is not added with :py:meth:`.add_data_part`
        :return: list of errors during write.
        """
        if calculation.measurement_file_path not in self.file_dict:
            raise ValueError("Unknown measurement file")
        self.file_dict[calculation.measurement_file_path].dump_data()
        return self.file_dict[calculation.measurement_file_path].get_errors()



import typing
from dataclasses import dataclass

import numpy as np

from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis.algorithm_description import AnalysisAlgorithmSelection
from PartSegCore.analysis.analysis_utils import SegmentationPipeline
from PartSegCore.mask_create import calculate_mask
from PartSegCore.project_info import HistoryElement
from PartSegCore.roi_info import ROIInfo
from PartSegCore.segmentation.algorithm_base import AdditionalLayerDescription, ROIExtractionResult
from PartSegImage import Image

if typing.TYPE_CHECKING:
    from PartSegCore.segmentation import RestartableAlgorithm


def _empty_fun(_a1, _a2):
    """Empty function for calculation run callback"""


@dataclass(frozen=True)
class PipelineResult:
    roi_info: ROIInfo
    additional_layers: dict[str, AdditionalLayerDescription]
    mask: np.ndarray
    history: list[HistoryElement]
    description: str


def calculate_pipeline(image: Image, mask: typing.Optional[np.ndarray], pipeline: SegmentationPipeline, report_fun):
    history = []
    report_fun("max", 2 * len(pipeline.mask_history) + 1)
    for i, el in enumerate(pipeline.mask_history):
        result, _ = calculate_segmentation_step(el.segmentation, image, mask)
        roi_info = result.roi_info.fit_to_image(image)
        report_fun("step", 2 * i + 1)
        new_mask = calculate_mask(
            mask_description=el.mask_property,
            roi=roi_info.roi,
            old_mask=mask,
            spacing=image.spacing,
            time_axis=image.time_pos,
        )
        segmentation_parameters = {"algorithm_name": el.segmentation.name, "values": el.segmentation.values}
        history.append(HistoryElement.create(roi_info, mask, segmentation_parameters, el.mask_property))
        report_fun("step", 2 * i + 2)
        mask = image.fit_array_to_image(new_mask)
    result, text = calculate_segmentation_step(pipeline.segmentation, image, mask)
    report_fun("step", 2 * len(pipeline.mask_history) + 1)
    return PipelineResult(result.roi_info, result.additional_layers, mask, history, text)


def calculate_segmentation_step(
    profile: ROIExtractionProfile, image: Image, mask: typing.Optional[np.ndarray]
) -> tuple[ROIExtractionResult, str]:
    algorithm: RestartableAlgorithm = AnalysisAlgorithmSelection[profile.algorithm]()
    algorithm.set_image(image)
    algorithm.set_mask(mask)
    parameters = profile.values
    algorithm.set_parameters(parameters)
    return algorithm.calculation_run(_empty_fun), algorithm.get_info_text()



import napari
from magicgui.widgets import create_widget
from napari.layers import Image as NapariImage
from qtpy.QtWidgets import QVBoxLayout, QWidget

from PartSeg.common_gui.dict_viewer import DictViewer


class LayerMetadata(QWidget):
    def __init__(self, viewer: napari.Viewer):
        super().__init__()
        self._viewer = viewer
        self.layer_selector = create_widget(annotation=NapariImage, label="Layer", options={})
        self._dict_viewer = DictViewer()
        layout = QVBoxLayout()
        layout.addWidget(self.layer_selector.native)
        layout.addWidget(self._dict_viewer)

        self.setLayout(layout)
        self.update_metadata()
        self.layer_selector.changed.connect(self.update_metadata)

    def reset_choices(self):
        self.layer_selector.reset_choices()

    def update_metadata(self):
        if self.layer_selector.value is None:
            return
        self._dict_viewer.set_data(self.layer_selector.value.metadata)



import json
import tarfile
import typing
from io import BufferedIOBase, BytesIO, IOBase, RawIOBase, TextIOBase
from pathlib import Path

import numpy as np

from PartSegCore.analysis.io_utils import ProjectTuple
from PartSegCore.io_utils import LoadBase
from PartSegCore.roi_info import ROIInfo
from PartSegCore.segmentation.noise_filtering import DimensionType
from PartSegCore.universal_const import UNIT_SCALE, Units
from PartSegImage import Image


class LoadPartSegOld(LoadBase):
    @classmethod
    def get_name(cls):
        return "Project old (*.tgz *.tbz2 *.gz *.bz2)"

    @classmethod
    def get_short_name(cls):
        return "project_old"

    @classmethod
    def _load(cls, tar_file: tarfile.TarFile, file_path: str) -> ProjectTuple:
        image_buffer = BytesIO()
        image_tar = tar_file.extractfile(tar_file.getmember("image.npy"))
        image_buffer.write(image_tar.read())
        image_buffer.seek(0)
        image_arr = np.load(image_buffer)
        try:
            res_buffer = BytesIO()
            res_tar = tar_file.extractfile(tar_file.getmember("res_mask.npy"))
            res_buffer.write(res_tar.read())
            res_buffer.seek(0)
            seg_array = np.load(res_buffer)
        except KeyError:
            seg_array = None
        algorithm_dict = json.load(tar_file.extractfile("data.json"))
        spacing = np.array(algorithm_dict["spacing"][::-1]) / UNIT_SCALE[Units.nm.value]
        image = Image(image_arr.reshape((1, *image_arr.shape, 1)), spacing, file_path, axes_order="TZYXC")
        values = {
            "channel": 0,
            "minimum_size": algorithm_dict["minimum_size"],
            "threshold": {"name": "Manual", "values": {"threshold": algorithm_dict["threshold"]}},
            "noise_removal": (
                {"name": "Gauss", "values": {"gauss_type": DimensionType.Layer, "radius": 1.0}}
                if algorithm_dict["use_gauss"]
                else {"name": "None", "values": {}}
            ),
            "side_connection": True,
        }

        algorithm_parameters = {
            "name": "Upper threshold" if algorithm_dict["threshold_type"] == "Upper" else "Lower threshold",
            "values": values,
            "threshold_list": algorithm_dict["threshold_list"],
        }
        return ProjectTuple(file_path, image, ROIInfo(seg_array), algorithm_parameters=algorithm_parameters)

    # noinspection DuplicatedCode
    @classmethod
    def load(
        cls,
        load_locations: list[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Optional[typing.Callable[[int, int], typing.Any]] = None,
        step_changed: typing.Optional[typing.Callable[[int], typing.Any]] = None,
        metadata: typing.Optional[dict] = None,
    ):
        """Load project from archive old format"""
        file_ob: typing.Union[str, tarfile.TarFile, TextIOBase, BufferedIOBase, RawIOBase, IOBase] = load_locations[0]
        if isinstance(file_ob, tarfile.TarFile):
            res = cls._load(file_ob, "")
        elif isinstance(file_ob, str):
            with tarfile.open(file_ob) as tar_file:
                res = cls._load(tar_file, file_ob)
        elif isinstance(file_ob, (BufferedIOBase, RawIOBase, IOBase)):
            with tarfile.open(fileobj=file_ob) as tar_file:
                res = cls._load(tar_file, "")
        else:
            raise ValueError(f"wrong type of file_ argument: {type(file_ob)}")

        return res



"""
This module contains some not fully intuitive algorithm for splitting masked area.
They are designed for measurements, but should be also available as segmentation algorithm to
allow user to preview, for better intuition how it works and which parameters are good for their purpose.

Better option is to implement this utils as class based with base class :py:class:`AlgorithmDescribeBase`.
Then do not need to manage algorithm parameters in places where it is used.

Both class from this module are designed for spherical mask, but may be useful also for others.
"""

import typing

import numpy as np
import SimpleITK
from pydantic import Field
from scipy.ndimage import distance_transform_edt

from PartSegCore.algorithm_describe_base import AlgorithmDescribeBase
from PartSegCore.universal_const import UNIT_SCALE, Units
from PartSegCore.utils import BaseModel


class BorderRimParameters(BaseModel):
    distance: float = Field(500, ge=0, le=10**6)
    units: Units = Units.nm


class BorderRim(AlgorithmDescribeBase):
    """
    This class implement border rim (Annulus like) calculation.
    For given mask and image spacing it marks as 1 all pixels in given distance from mask border.

    https://en.wikipedia.org/wiki/Annulus_(mathematics)

    The algorithm is:

    1. For each image voxel calculate distance from background (0 labeled voxels in mask) with respect of voxel size
    2. Select these voxels which are closer than given distance.
    """

    __argument_class__ = BorderRimParameters

    @classmethod
    def get_name(cls) -> str:
        return "Border Rim"

    @staticmethod
    def border_mask(mask: np.ndarray, distance: float, units: Units, voxel_size, **_) -> typing.Optional[np.ndarray]:
        """
        This is function which implement calculation.

        :param mask: area for which rim should be calculated. 2d or 3d numpy array,
        :param distance: distance from border which will be marked.
        :param units: in which unit distance is given
        :param voxel_size: Image spacing in absolute units
        :param _: ignored arguments
        :return: border rim marked with 1
        """
        if mask is None:
            return None
        units_scalar = UNIT_SCALE[units.value]
        final_radius = [int((distance / units_scalar) / x) for x in reversed(voxel_size)]
        mask = np.array(mask > 0)
        mask = mask.astype(np.uint8)
        eroded = SimpleITK.GetArrayFromImage(
            SimpleITK.BinaryErode(SimpleITK.GetImageFromArray(mask.squeeze()), final_radius)
        )
        eroded = eroded.reshape(mask.shape)
        mask[eroded > 0] = 0
        return mask


class MaskDistanceSplitParameters(BaseModel):
    num_of_parts: int = Field(2, title="Number of Parts", ge=1, le=1024)
    equal_volume: bool = Field(
        False,
        title="Equal Volume",
        description="If split should be done in respect of parts volume of parts thickness.",
    )


class MaskDistanceSplit(AlgorithmDescribeBase):
    """
    This class contains implementation of splitting mask on parts based on distance from borders.
    Has two modes or working. It may split on parts with same thickness or same volume..

    Flow chart of algorithm:

    .. graphviz::

       digraph model {
          "Mask" -> "Voxel distance from background";
          "Voxel distance from background" -> "maximum distance"[label=<Equal thickness>]
          "maximum distance" -> "Bounds" -> "Split voxels in respect of bounds"
          "Bounds"[label=<Bounds<br />
          <FONT POINT-SIZE="10">[0, (maximum distance)/(Number of Parts),<br />
          2*(maximum distance)/(Number of Parts)<br />, ... ,maximum distance]</FONT>>]
          "Voxel distance from background" -> "Distance histogram"[label=<Equal volume>]
          "Distance histogram" -> "Create bounds base on histogram" -> "Split voxels in respect of bounds"
          "Create bounds base on histogram"[label=<Create bounds base on histogram<br />
          <FONT POINT-SIZE="10">using histogram bins approximate equal number of voxels in each part</FONT>>]
          "Distance histogram"[label=<Distance histogram<br /><font point-size="10">of voxels with positive distance
          <br /> from background </font>>]

       }

    """

    __argument_class__ = MaskDistanceSplitParameters

    @classmethod
    def get_name(cls) -> str:
        return "Mask Distance Split"

    @staticmethod
    def split(mask: np.ndarray, num_of_parts: int, equal_volume: bool, voxel_size, **_):
        """
        This is function which implement calculation.

        :param mask: area for which rim should be calculated. 2d or 3d numpy array
        :param num_of_parts: num of parts on which mask should be split
        :param equal_volume: if split should be on equal volume or equal thick
        :param voxel_size: image voxel size
        :return: mask region labelled starting from 1 near border
        """
        if len(voxel_size) == 2 and mask.ndim == 3:
            voxel_size = (1, *voxel_size)
        distance_arr = distance_transform_edt(mask, sampling=voxel_size)
        if equal_volume:
            # TODO add more bins, fix tests for more bins
            hist, bins = np.histogram(distance_arr[distance_arr > 0], bins=10 * num_of_parts)
            total = np.sum(hist)
            levels, step = np.linspace(0, total, num_of_parts + 1, True, retstep=True)
            bounds = [0]
            i = 1
            cum_sum = 0
            for val, begin, end in zip(hist, bins, bins[1:]):
                cum_sum += val
                if cum_sum > levels[i]:
                    exceed = (cum_sum - levels[i]) / step
                    bounds.append(begin + (end - begin) * exceed)
                    i += 1
        else:
            max_dist = np.max(distance_arr)
            bounds = np.linspace(0, max_dist, num_of_parts, False)
        mask = np.zeros(mask.shape, dtype=np.uint8 if num_of_parts < 255 else np.uint16)
        for bound in bounds:
            mask[distance_arr > bound] += 1
        return mask



import warnings
from abc import ABC

import numpy as np

from PartSegCore.algorithm_describe_base import AlgorithmDescribeBase, AlgorithmProperty, AlgorithmSelection


class BaseMuMid(AlgorithmDescribeBase, ABC):
    @classmethod
    def get_fields(cls):
        return []

    @classmethod
    def value(cls, sprawl_area: np.ndarray, data: np.ndarray, lower_bound, upper_bound, arguments: dict):
        raise NotImplementedError


class MeanBound(BaseMuMid):
    @classmethod
    def get_name(cls):
        return "Mean bound"

    @classmethod
    def value(cls, sprawl_area: np.ndarray, data: np.ndarray, lower_bound, upper_bound, arguments: dict):
        return (lower_bound + upper_bound) / 2


class PercentBound(BaseMuMid):
    @classmethod
    def get_name(cls):
        return "Percent from lower bound"

    @classmethod
    def get_fields(cls):
        return [
            AlgorithmProperty(
                "percent",
                "Percent",
                50,
                options_range=(0, 100),
                help_text="Calculate: lower_value + (upper_bound - lower_bound) * percent / 100",
            )
        ]

    @classmethod
    def value(cls, sprawl_area: np.ndarray, data: np.ndarray, lower_bound, upper_bound, arguments: dict):
        return min(lower_bound, upper_bound) + abs(lower_bound - upper_bound) * arguments["percent"] / 100


class MeanPixelValue(BaseMuMid):
    @classmethod
    def get_name(cls):
        return "Mean pixel value"

    @classmethod
    def value(cls, sprawl_area: np.ndarray, data: np.ndarray, lower_bound, upper_bound, arguments: dict):
        return np.mean[data[sprawl_area > 0]]


class MedianPixelValue(BaseMuMid):
    @classmethod
    def get_name(cls):
        return "Median pixel value"

    @classmethod
    def value(cls, sprawl_area: np.ndarray, data: np.ndarray, lower_bound, upper_bound, arguments: dict):
        return np.median[data[sprawl_area > 0]]


class QuantilePixelValue(BaseMuMid):
    @classmethod
    def get_name(cls):
        return "Quantile pixel value"

    @classmethod
    def get_fields(cls):
        return [AlgorithmProperty("quantile", "Quantile", 50, options_range=(0, 100))]

    @classmethod
    def value(cls, sprawl_area: np.ndarray, data: np.ndarray, lower_bound, upper_bound, arguments: dict):
        return np.quantile[data[sprawl_area > 0], arguments["quantile"] / 100]


class MuMidSelection(AlgorithmSelection, class_methods=["value"], suggested_base_class=BaseMuMid):
    pass


MuMidSelection.register(MeanBound)
MuMidSelection.register(PercentBound)
MuMidSelection.register(MeanPixelValue)
MuMidSelection.register(MedianPixelValue)
MuMidSelection.register(QuantilePixelValue)


def __getattr__(name):  # pragma: no cover
    if name == "mu_mid_dict":
        warnings.warn(
            "mu_mid_dict is deprecated. Please use MuMidSelection instead", category=FutureWarning, stacklevel=2
        )
        return MuMidSelection.__register__
    raise AttributeError(f"module {__name__} has no attribute {name}")



from typing import Any

import pytest
from local_migrator import REGISTER, class_to_str, register_class, rename_key, update_argument


@register_class
class SampleClass1:
    pass


def rename_a_to_c(dkt: dict[str, Any]) -> dict[str, Any]:
    dkt = dict(dkt)
    dkt["c"] = dkt["a"]
    del dkt["a"]
    return dkt


@register_class(version="0.0.1", migrations=[("0.0.1", rename_a_to_c)])
class SampleClass2:
    pass


class SampleClass3:
    pass


@register_class(old_paths=["test.test.BBase"], version="0.0.2")
class SampleClass4:
    pass


class SampleClass6:
    pass


def test_migrate():
    assert REGISTER.migrate_data(class_to_str(SampleClass1), {}, {"a": 1, "b": 2}) == {"a": 1, "b": 2}
    assert REGISTER.migrate_data(
        class_to_str(SampleClass2), {class_to_str(SampleClass2): "0.0.1"}, {"a": 1, "b": 2}
    ) == {"a": 1, "b": 2}
    assert REGISTER.migrate_data(class_to_str(SampleClass2), {}, {"a": 1, "b": 2}) == {"c": 1, "b": 2}


def test_unregistered_class():
    assert REGISTER.get_class(class_to_str(SampleClass3)) is SampleClass3


def test_old_paths():
    assert REGISTER.get_class("test.test.BBase") is SampleClass4


def test_old_class_error():
    with pytest.raises(RuntimeError):
        register_class(SampleClass6, old_paths=[class_to_str(SampleClass4)])


def test_get_version():
    assert str(REGISTER.get_version(SampleClass1)) == "0.0.0"
    assert str(REGISTER.get_version(SampleClass2)) == "0.0.1"
    assert str(REGISTER.get_version(SampleClass4)) == "0.0.2"


def test_rename_key():
    dkt = {"aaa": 1, "bbb": 2}
    assert rename_key(from_key="aaa", to_key="ccc")(dkt) == {"bbb": 2, "ccc": 1}
    with pytest.raises(KeyError):
        rename_key("ccc", "ddd")(dkt)

    assert rename_key("ccc", "ddd", optional=True)(dkt) == dkt


def test_update_argument(clean_register):
    @REGISTER.register(version="0.0.1")
    class MigrateClass:
        def __init__(self, a, b):
            self.a = a
            self.b = b

    class ClassToCall:
        __argument_class__ = MigrateClass

        @classmethod
        @update_argument("arg")
        def call_func(cls, aa, arg: MigrateClass):
            assert arg.a == 1
            assert arg.b == 2
            assert aa == 1

    ClassToCall.call_func(aa=1, arg={"a": 1, "b": 2})
    ClassToCall.call_func(1, {"a": 1, "b": 2})


def test_migrate_parent_class(clean_register):
    @register_class(version="0.0.1", migrations=[("0.0.1", rename_key("field", "field1"))])
    class BaseMigrateClass:
        def __init__(self, field1):
            self.field1 = field1

    @register_class
    class MigrateClass(BaseMigrateClass):
        def __init__(self, field2, **kwargs):
            super().__init__(**kwargs)
            self.field2 = field2

    migrated = REGISTER.migrate_data(class_to_str(MigrateClass), {}, {"field2": 1, "field": 5})
    assert "field1" in migrated



import typing

import numpy as np

from PartSeg.common_backend.progress_thread import ProgressTread
from PartSegCore.analysis.analysis_utils import SegmentationPipeline
from PartSegCore.analysis.calculate_pipeline import PipelineResult, calculate_pipeline
from PartSegImage import Image


class CalculatePipelineThread(ProgressTread):
    result: typing.Optional[PipelineResult]

    def __init__(self, image: Image, mask: typing.Union[np.ndarray, None], pipeline: SegmentationPipeline, parent=None):
        super().__init__(parent=parent)
        self.image = image
        self.mask = mask
        self.pipeline = pipeline
        self.result = None

    def run(self):
        try:
            self.result = calculate_pipeline(
                image=self.image, mask=self.mask, pipeline=self.pipeline, report_fun=self.info_function
            )
        except Exception as e:  # pylint: disable=broad-except
            self.error_signal.emit(e)



# pylint: disable=no-self-use

import itertools

import numpy as np
import pytest

from PartSegCore.segmentation.watershed import NeighType, calculate_distances_array
from PartSegCore_compiled_backend.multiscale_opening import MuType, PyMSO, calculate_mu, calculate_mu_mid
from PartSegCore_compiled_backend.sprawl_utils.euclidean_cython import calculate_euclidean


class TestMu:
    def test_base_mu(self):
        image = np.zeros((10, 10, 10), dtype=np.uint8)
        image[2:8, 2:8, 2:8] = 10
        res = calculate_mu(image, 2, 8, MuType.base_mu)
        assert np.all(res == (image > 0).astype(np.float64))
        res = calculate_mu(image, 5, 15, MuType.base_mu)
        assert np.all(res == (image > 0).astype(np.float64) * 0.5)
        image[4:6, 4:6, 4:6] = 20
        res = calculate_mu(image, 5, 15, MuType.base_mu)
        assert np.all(res == ((image > 0).astype(np.float64) + (image > 15).astype(np.float64)) * 0.5)

    def test_base_mu_masked(self):
        image = np.zeros((10, 10, 10), dtype=np.uint8)
        image[2:8, 2:8, 2:8] = 10
        res = calculate_mu(image, 2, 8, MuType.base_mu, image > 0)
        assert np.all(res == (image > 0).astype(np.float64))
        res = calculate_mu(image, 5, 15, MuType.base_mu, image > 0)
        assert np.all(res == (image > 0).astype(np.float64) * 0.5)
        image[4:6, 4:6, 4:6] = 20
        res = calculate_mu(image, 5, 15, MuType.base_mu, image > 0)
        assert np.all(res == ((image > 0).astype(np.float64) + (image > 15).astype(np.float64)) * 0.5)

    def test_reversed_base_mu(self):
        image = np.zeros((10, 10, 10), dtype=np.uint8)
        image[2:8, 2:8, 2:8] = 10
        res = calculate_mu(image, 8, 2, MuType.base_mu)
        assert np.all(res == (image == 0).astype(np.float64))
        res = calculate_mu(image, 15, 5, MuType.base_mu)
        assert np.all(res == ((20 - image) / 20).astype(np.float64))
        image[4:6, 4:6, 4:6] = 20
        res = calculate_mu(image, 15, 5, MuType.base_mu)
        assert np.all(res == (20 - image) / 20)

    def test_reversed_base_mu_masked(self):
        image = np.zeros((10, 10, 10), dtype=np.uint8)
        image[2:8, 2:8, 2:8] = 10
        mask = image > 0
        res = calculate_mu(image, 8, 2, MuType.base_mu, mask)
        assert np.all(res == np.zeros(image.shape, dtype=np.float64))
        res = calculate_mu(image, 15, 5, MuType.base_mu, mask)
        assert np.all(res == mask * 0.5)
        image[4:6, 4:6, 4:6] = 20
        res = calculate_mu(image, 15, 5, MuType.base_mu, mask)
        assert np.all(res == (mask * (image < 20)) * 0.5)

    def test_reflection_mu(self):
        image = np.zeros((10, 10, 10), dtype=np.uint8)
        image[2:8, 2:8, 2:8] = 10
        res = calculate_mu(image, 2, 8, MuType.reflection_mu)
        assert np.all(res == np.ones(image.shape, dtype=np.float64))
        res = calculate_mu(image, 5, 15, MuType.reflection_mu)
        assert np.all(res == (20 - image) / 20)
        image[4:6, 4:6, 4:6] = 20
        res = calculate_mu(image, 5, 15, MuType.reflection_mu)
        assert np.all(res == np.ones(image.shape, dtype=np.float64) - (image == 10) * 0.5)

    def test_reflection_mu_masked(self):
        image = np.zeros((10, 10, 10), dtype=np.uint8)
        image[2:8, 2:8, 2:8] = 10
        mask = image > 0
        res = calculate_mu(image, 2, 8, MuType.reflection_mu, mask)
        assert np.all(res == mask * 1.0)
        res = calculate_mu(image, 5, 15, MuType.reflection_mu, mask)
        assert np.all(res == ((20 - image) / 20) * mask)
        image[4:6, 4:6, 4:6] = 20
        res = calculate_mu(image, 5, 15, MuType.reflection_mu, mask)
        assert np.all(res == (np.ones(image.shape, dtype=np.float64) - (image == 10) * 0.5) * mask)

    def test_reversed_reflection_mu(self):
        image = np.zeros((10, 10, 10), dtype=np.uint8)
        image[2:8, 2:8, 2:8] = 10
        res = calculate_mu(image, 8, 2, MuType.reflection_mu)
        assert np.all(res == np.ones(image.shape, dtype=np.float64))
        res = calculate_mu(image, 15, 5, MuType.reflection_mu)
        assert np.all(res == (20 - image) / 20)
        image[4:6, 4:6, 4:6] = 20
        res = calculate_mu(image, 15, 5, MuType.reflection_mu)
        assert np.all(res == np.ones(image.shape, dtype=np.float64) - (image == 10) * 0.5)

    def test_reversed_reflection_mu_masked(self):
        image = np.zeros((10, 10, 10), dtype=np.uint8)
        image[2:8, 2:8, 2:8] = 10
        mask = image > 0
        res = calculate_mu(image, 8, 2, MuType.reflection_mu, mask)
        assert np.all(res == mask * 1.0)
        res = calculate_mu(image, 15, 5, MuType.reflection_mu, mask)
        assert np.all(res == ((20 - image) / 20) * mask)
        image[4:6, 4:6, 4:6] = 20
        res = calculate_mu(image, 15, 5, MuType.reflection_mu, mask)
        assert np.all(res == (np.ones(image.shape, dtype=np.float64) - (image == 10) * 0.5) * mask)

    def test_two_object_mu(self):
        """
        TODO add this test
        """

    def test_reshape(self):
        image = np.zeros((40, 150, 120), dtype=np.uint16)
        image[2:-2, 2:-2, 2:-2] = 30
        res = calculate_mu(image, 20, 40, MuType.base_mu)
        assert np.all(res == (image > 0) * 0.5)


def test_mso_construct():
    data = PyMSO()
    assert hasattr(data, "set_image")


class TestConstrainedDilation:
    def test_two_components_base(self):
        components = np.zeros((10, 10, 20), dtype=np.uint8)
        components[4:6, 4:6, 4:6] = 1
        components[4:6, 4:6, 14:16] = 2
        sprawl_area = np.zeros(components.shape, dtype=np.uint8)
        sprawl_area[2:8, 2:8, 2:18] = True
        sprawl_area[components > 0] = False
        fdt = np.zeros(components.shape, dtype=np.float64)
        mso = PyMSO()
        neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
        mso.set_neighbourhood(neigh, dist)
        mso.set_components(components)
        mso.set_mu_array(np.ones(components.shape))
        res = mso.constrained_dilation(fdt, components, sprawl_area)
        assert np.all(res == components)

        fdt[:] = 20
        components2 = np.zeros(components.shape, dtype=np.uint8)
        components2[2:8, 2:8, 2:10] = 1
        components2[2:8, 2:8, 10:18] = 2
        res = mso.constrained_dilation(fdt, components, sprawl_area)
        assert np.all(res == components2)

    def test_low_fdt_two_components(self):
        components = np.zeros((10, 10, 20), dtype=np.uint8)
        components[4:6, 4:6, 4:6] = 1
        components[4:6, 4:6, 14:16] = 2
        sprawl_area = np.zeros(components.shape, dtype=np.uint8)
        sprawl_area[2:8, 2:8, 2:18] = True
        sprawl_area[components > 0] = False
        fdt = np.ones(components.shape, dtype=np.float64) * 20
        mso = PyMSO()
        neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
        mso.set_neighbourhood(neigh, dist)
        mso.set_components(components)
        mso.set_mu_array(np.ones(components.shape))
        components2 = np.zeros(components.shape, dtype=np.uint8)
        components2[2:8, 2:8, 2:10] = 1
        components2[2:8, 2:8, 10:18] = 2
        fdt2 = np.copy(fdt)
        fdt2[:, :, 7] = 2
        components3 = np.copy(components2)
        components3[2:8, 2:8, 8:18] = 2
        components3[2:8, 2:8, 7] = 0
        res = mso.constrained_dilation(fdt2, components, sprawl_area)
        assert np.all(res == components3)
        fdt2 = np.copy(fdt)
        fdt2[:, :, 12] = 2
        components3 = np.copy(components2)
        components3[2:8, 2:8, 2:12] = 1
        components3[2:8, 2:8, 12] = 0
        res = mso.constrained_dilation(fdt2, components, sprawl_area)
        assert np.all(res == components3)
        fdt2 = np.copy(fdt)
        fdt2[:, :, 12] = 2
        fdt2[:, :, 7] = 2
        components3 = np.copy(components2)
        components3[2:8, 2:8, 7:13] = 0
        res = mso.constrained_dilation(fdt2, components, sprawl_area)
        assert np.all(res == components3)

    def test_high_fdt_two_components(self):
        components = np.zeros((10, 10, 20), dtype=np.uint8)
        components[4:6, 4:6, 4:6] = 1
        components[4:6, 4:6, 14:16] = 2
        sprawl_area = np.zeros(components.shape, dtype=np.uint8)
        sprawl_area[2:8, 2:8, 2:18] = True
        sprawl_area[components > 0] = False
        fdt = np.ones(components.shape, dtype=np.float64) * 20
        mso = PyMSO()
        neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
        mso.set_neighbourhood(neigh, dist)
        mso.set_components(components)
        mso.set_mu_array(np.ones(components.shape))
        components2 = np.zeros(components.shape, dtype=np.uint8)
        components2[2:8, 2:8, 2:10] = 1
        components2[2:8, 2:8, 10:18] = 2
        fdt2 = np.copy(fdt)
        fdt2[:, :, 7] = 25
        components3 = np.copy(components2)
        components3[2:8, 2:8, 8:18] = 2
        components3[2:8, 2:8, 7] = 0
        res = mso.constrained_dilation(fdt2, components, sprawl_area)
        assert np.all(res == components3)
        fdt2 = np.copy(fdt)
        fdt2[:, :, 12] = 25
        components3 = np.copy(components2)
        components3[2:8, 2:8, 2:12] = 1
        components3[2:8, 2:8, 12] = 0
        res = mso.constrained_dilation(fdt2, components, sprawl_area)
        assert np.all(res == components3)
        fdt2 = np.copy(fdt)
        fdt2[:, :, 12] = 25
        fdt2[:, :, 7] = 25
        components3 = np.copy(components2)
        components3[2:8, 2:8, 7:13] = 0
        res = mso.constrained_dilation(fdt2, components, sprawl_area)
        assert np.all(res == components3)

    def test_chain_component(self):
        for i in range(2, 10):
            components = np.zeros((10, 10, i * 10), dtype=np.uint8)
            for j in range(i):
                components[4:6, 4:6, (10 * j + 4) : (10 * j + 6)] = j + 1
            fdt = np.ones(components.shape, dtype=np.float64) * i * 10
            sprawl_area = np.zeros(components.shape, dtype=np.uint8)
            sprawl_area[2:8, 2:8, 2 : (10 * i) - 2] = True
            sprawl_area[components > 0] = False
            components2 = np.zeros(components.shape, dtype=np.uint8)
            for j in range(i):
                components2[2:8, 2:8, (j * 10) : (j + 1) * 10] = j + 1
            components2[:, :, :2] = 0
            components2[:, :, -2:] = 0
            mso = PyMSO()
            neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
            mso.set_neighbourhood(neigh, dist)
            mso.set_components(components)
            mso.set_mu_array(np.ones(components.shape))
            res = mso.constrained_dilation(fdt, components, sprawl_area)
            assert np.all(res == components2)

    def test_chain_component_with_wall(self):
        for i in range(2, 10):
            components = np.zeros((10, 10, i * 10), dtype=np.uint8)
            for j in range(i):
                components[4:6, 4:6, (10 * j + 4) : (10 * j + 6)] = j + 1
            fdt = np.ones(components.shape, dtype=np.float64) * i * 10
            sprawl_area = np.zeros(components.shape, dtype=np.uint8)
            sprawl_area[2:8, 2:8, 2 : (10 * i) - 2] = True
            sprawl_area[components > 0] = False
            components2 = np.zeros(components.shape, dtype=np.uint8)
            for j in range(i):
                components2[2:8, 2:8, (j * 10 + 1) : (j + 1) * 10] = j + 1
                fdt[:, :, j * 10] = i * 10 + 2
            components2[:, :, :2] = 0
            components2[:, :, -2:] = 0
            mso = PyMSO()
            neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
            mso.set_neighbourhood(neigh, dist)
            mso.set_components(components)
            mso.set_mu_array(np.ones(components.shape))
            res = mso.constrained_dilation(fdt, components, sprawl_area)
            assert np.all(res == components2)


class TestOptimumErosionCalculate:
    def test_two_components(self):
        components = np.zeros((10, 10, 20), dtype=np.uint8)
        components[4:6, 4:6, 4:6] = 1
        components[4:6, 4:6, 14:16] = 2
        sprawl_area = np.zeros(components.shape, dtype=np.uint8)
        sprawl_area[2:8, 2:8, 2:18] = True
        sprawl_area[components > 0] = False
        fdt = np.zeros(components.shape, dtype=np.float64)
        mso = PyMSO()
        neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
        mso.set_neighbourhood(neigh, dist)
        mso.set_components(components)
        res = mso.optimum_erosion_calculate(fdt, components, sprawl_area)
        assert np.all(res == components)

        fdt[:] = 2
        fdt[:, :, 10] = 1
        components2 = np.zeros(components.shape, dtype=np.uint8)
        components2[2:8, 2:8, 2:10] = 1
        components2[2:8, 2:8, 11:18] = 2
        res = mso.optimum_erosion_calculate(fdt, components, sprawl_area)
        assert np.all(res == components2)
        fdt[5, 5, 10] = 2
        res = mso.optimum_erosion_calculate(fdt, components, sprawl_area)
        assert np.all(res == components)

    def test_chain_component_base(self):
        for i in range(2, 10):
            components = np.zeros((4, 5, i * 5), dtype=np.uint8)
            for j in range(i):
                components[2, 2, 5 * j + 2] = j + 1
            fdt = np.ones(components.shape, dtype=np.float64) * 2
            for j in range(i - 1):
                fdt[:, :, j * 5 + 4] = 1
            sprawl_area = np.ones(components.shape, dtype=np.uint8)
            sprawl_area[components > 0] = False
            components2 = np.zeros(components.shape, dtype=np.uint8)
            for j in range(i):
                components2[:, :, (j * 5) : (j * 5) + 4] = j + 1
            components2[:, :, -1] = i
            neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
            mso = PyMSO()
            mso.set_neighbourhood(neigh, dist)
            mso.set_components(components)
            res = mso.optimum_erosion_calculate(fdt, components, sprawl_area)
            assert np.all(res == components2)

    def test_chain_component(self):
        for i in range(2, 10):
            components = np.zeros((10, 10, i * 10), dtype=np.uint8)
            for j in range(i):
                components[4:6, 4:6, (10 * j + 4) : (10 * j + 6)] = j + 1
            fdt = np.ones(components.shape, dtype=np.float64) * 2
            for j in range(i - 1):
                fdt[:, :, (j + 1) * 10] = 1
            sprawl_area = np.zeros(components.shape, dtype=np.uint8)
            sprawl_area[2:8, 2:8, 2 : (10 * i) - 2] = True
            sprawl_area[components > 0] = False
            components2 = np.zeros(components.shape, dtype=np.uint8)
            for j in range(i):
                components2[2:8, 2:8, (j * 10 + 1) : (j + 1) * 10] = j + 1
            components2[:, :, :2] = 0
            components2[:, :, -2:] = 0
            neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
            mso = PyMSO()
            mso.set_neighbourhood(neigh, dist)
            mso.set_components(components)
            res = mso.optimum_erosion_calculate(fdt, components, sprawl_area)
            assert np.all(res == components2)

    def test_bridges(self):
        components = np.zeros((10, 20, 30), dtype=np.uint8)
        components[4:6, 4:6, 4:6] = 1
        components[4:6, 4:6, 24:26] = 2
        components[4:6, 14:16, 14:16] = 3
        sprawl_area = np.zeros(components.shape, dtype=np.uint8)
        sprawl_area[2:9, 2:10, 2:28] = True
        sprawl_area[2:9, 10:18, 11:20] = True
        sprawl_area[components > 0] = False
        fdt = np.ones(components.shape, dtype=np.float64) * 2
        fdt[:, 10, :] = 1
        fdt[:, :, 10] = 1
        fdt[:, :, 20] = 1

        comp = np.copy(components)
        comp[2:9, 2:10, 2:10] = 1
        comp[2:9, 2:10, 21:28] = 2
        comp[2:9, 11:18, 11:20] = 3
        neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
        mso = PyMSO()
        mso.set_neighbourhood(neigh, dist)
        mso.set_components(components)
        res = mso.optimum_erosion_calculate(fdt, components, sprawl_area)
        assert np.all(res == comp)
        fdt2 = np.copy(fdt)
        fdt2[5, 5, 10] = 2
        comp2 = np.copy(comp)
        comp2[5, 5, 10] = 1
        comp2[2:9, 2:10, 11:20] = 1
        res = mso.optimum_erosion_calculate(fdt2, components, sprawl_area)
        assert np.all(res == comp2)
        fdt2 = np.copy(fdt)
        fdt2[5, 5, 20] = 2
        comp2 = np.copy(comp)
        comp2[5, 5, 20] = 2
        comp2[2:9, 2:10, 11:20] = 2
        res = mso.optimum_erosion_calculate(fdt2, components, sprawl_area)
        assert np.all(res == comp2)
        fdt2 = np.copy(fdt)
        fdt2[5, 10, 15] = 2
        comp2 = np.copy(comp)
        comp2[5, 10, 15] = 3
        comp2[2:9, 2:10, 11:20] = 3
        res = mso.optimum_erosion_calculate(fdt2, components, sprawl_area)
        assert np.all(res == comp2)


class TestFDT:
    def test_fdt_base(self):
        mso = PyMSO()
        neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
        components = np.zeros((10, 10, 10), dtype=np.uint8)
        components[3:7, 3:7, 3:7] = 1
        mso.set_neighbourhood(neigh, dist)
        mso.set_components(components)
        mso.set_mu_array(np.ones(components.shape))
        res = mso.calculate_FDT()
        arr = calculate_euclidean((components == 0).astype(np.uint8), components, neigh, dist)
        assert np.all(res == arr)
        mso.set_mu_array(np.ones(components.shape) * 0.5)
        res = mso.calculate_FDT()
        assert np.all(res == arr * 0.5)
        mu = np.ones(components.shape) * 0.5
        mu[components > 0] = 1
        mso.set_mu_array(mu)
        arr *= 0.5
        arr[arr > 0] += 0.25
        arr[3:7, (0, 1, 2, 0, 1, 2, 9, 8, 7, 9, 8, 7), (0, 1, 2, 9, 8, 7, 0, 1, 2, 9, 8, 7)] += np.sqrt(2) / 4 - 0.25
        for i in range(3):
            lb = i
            ub = 9 - i
            arr[lb, lb + 1 : ub, (lb, ub)] += np.sqrt(2) / 4 - 0.25
            arr[lb, (lb, ub), lb + 1 : ub] += np.sqrt(2) / 4 - 0.25
            arr[ub, lb + 1 : ub, (lb, ub)] += np.sqrt(2) / 4 - 0.25
            arr[ub, (lb, ub), lb + 1 : ub] += np.sqrt(2) / 4 - 0.25
            for el in itertools.product([lb, ub], repeat=3):
                arr[el] += np.sqrt(3) / 4 - 0.25
        for z, (y, x) in itertools.product([2, 7], itertools.product([0, 9], repeat=2)):
            arr[z, y, x] += np.sqrt(2) / 4 - 0.25
        for z, (y, x) in itertools.product([2, 7], itertools.product([1, 8], repeat=2)):
            arr[z, y, x] += np.sqrt(2) / 4 - 0.25
        for z, (y, x) in itertools.product([1, 8], itertools.product([0, 9], repeat=2)):
            arr[z, y, x] += np.sqrt(2) / 4 - 0.25
        res2 = mso.calculate_FDT()
        assert np.allclose(res2, arr)

    def test_fdt_simple(self):
        mso = PyMSO()
        neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
        components = np.zeros((3, 3, 3), dtype=np.uint8)
        components[1, 1, 1] = 1
        mso.set_neighbourhood(neigh, dist)
        mso.set_components(components)
        mso.set_mu_array(np.ones(components.shape))
        arr = np.zeros(components.shape)
        arr[(0, 0, 0, 0, 2, 2, 2, 2), (0, 0, 2, 2, 0, 0, 2, 2), (0, 2, 0, 2, 0, 2, 0, 2)] = np.sqrt(3)
        arr[
            (0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2),
            (0, 1, 1, 2, 0, 0, 2, 2, 0, 1, 1, 2),
            (1, 0, 2, 1, 0, 2, 0, 2, 1, 0, 2, 1),
        ] = np.sqrt(2)
        arr[(0, 1, 1, 1, 1, 2), (1, 0, 1, 1, 2, 1), (1, 1, 0, 2, 1, 1)] = 1
        res = mso.calculate_FDT()
        assert np.all(res == arr)


class TestExceptions:
    def test_fdt(self):
        mso = PyMSO()
        neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
        components = np.zeros((10, 10, 10), dtype=np.uint8)
        components[3:7, 3:7, 3:7] = 1
        mso.set_neighbourhood(neigh, dist)
        with pytest.raises(RuntimeError):
            mso.calculate_FDT()
        mso.set_components(components)
        with pytest.raises(RuntimeError):
            mso.calculate_FDT()


class TestMSO:
    def test_two_components(self):
        components = np.zeros((10, 10, 20), dtype=np.uint8)
        components[:] = 1
        components[2:8, 2:8, 2:18] = 0
        components[4:6, 4:6, 4:6] = 2
        components[4:6, 4:6, 14:16] = 3
        mu_arr = np.zeros(components.shape, dtype=np.float64)
        mu_arr[components == 0] = 0.5
        mu_arr[components > 1] = 1

        mso = PyMSO()
        neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
        mso.set_neighbourhood(neigh, dist)
        mso.set_components(components)
        mso.set_mu_array(mu_arr)
        mso.set_components_num(3)

        mso.run_MSO(10)
        mso.steps_done()
        arr = np.copy(components)
        arr[arr == 1] = 0
        arr[3:7, 3:7, 3:10] = 2
        arr[3:7, 3:7, 10:17] = 3

        mu_arr[2:8, 2:8, 10] = 0.08
        mso.set_mu_array(mu_arr)
        mso.run_MSO(10)
        arr[2:8, 2:8, 2:10] = 2
        arr[2:8, 2:8, 11:18] = 3
        arr[3:7, 3:7, 10] = 0
        mu_arr[2:8, 2:8, 9] = 0.08
        mso.set_mu_array(mu_arr)
        arr[2:8, 2:8, 2:9] = 2
        arr[2:8, 2:8, 9] = 0
        arr[2:8, 2:8, 11:18] = 3
        mso.run_MSO(10)
        res = mso.get_result_catted()
        assert np.all(arr == res)

    def test_two_components_bridge(self):
        components = np.zeros((10, 10, 20), dtype=np.uint8)
        components[:] = 1
        components[2:8, 2:8, 2:18] = 0
        components[4:6, 4:6, 4:6] = 2
        components[4:6, 4:6, 14:16] = 3
        mu_arr = np.zeros(components.shape, dtype=np.float64)
        mu_arr[components == 0] = 0.5
        mu_arr[components > 1] = 1
        mu_arr[2:8, 2:8, (9, 10)] = 0.1
        mso = PyMSO()
        neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
        mso.set_neighbourhood(neigh, dist)
        mso.set_components(components, 3)
        mso.set_mu_array(mu_arr)
        mso.set_components_num(3)
        mso.run_MSO(10)
        res = mso.get_result_catted()
        assert mso.steps_done() == 2
        arr = np.copy(components)
        arr[arr == 1] = 0
        arr[2:8, 2:8, 2:9] = 2
        arr[2:8, 2:8, 11:18] = 3
        assert np.all(res == arr)

    def test_background_simple(self):
        components = np.ones((20, 20, 20), dtype=np.uint8)
        components[1:-1, 1:-1, 1:-1] = 0
        components[9:11, 9:11, 9:11] = 2

        mu_array = np.zeros(components.shape)
        mu_array[1:-1, 1:-1, 1:-1] = 0.7
        mu_array[3:-3, 3:-3, 3:-3] = 0.6
        mu_array[5:-5, 5:-5, 5:-5] = 0.4
        mu_array[6:-6, 6:-6, 6:-6] = 0.6
        mu_array[8:-8, 8:-8, 8:-8] = 0.7
        mu_array[components > 0] = 1.0
        mso = PyMSO()
        neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
        mso.set_neighbourhood(neigh, dist)
        mso.set_use_background(True)
        mso.set_components(components, 3)
        mso.set_mu_array(mu_array)
        mso.set_components_num(3)
        mso.run_MSO(10)
        mso.get_result_catted()


class TestMuMid:
    def test_simple(self):
        data = np.zeros((10, 10, 10))
        data[1:-1, 1:-1, 1:-1] = 20
        ones = np.ones(data.shape)
        res = calculate_mu_mid(data, 5, 10, 15)
        assert np.all(ones == res)
        res = calculate_mu_mid(data, 5, 10, 20)
        assert np.all(ones == res)
        res = calculate_mu_mid(data, 5, 20, 30)
        assert np.all(res == (data == 0).astype(float))
        data[2:-2, 2:-2, 2:-2] = 30
        res = calculate_mu_mid(data, 5, 25, 30)
        assert np.all(res == (data != 20).astype(float) + (data == 20) * 0.25)
        res = calculate_mu_mid(data, 5, 25, 35)
        assert np.all(res == (data == 0).astype(float) + (data == 20) * 0.25 + (data == 30) * 0.5)


def test_calculate_distances_array():
    neigh, dist = calculate_distances_array((1, 1, 1), NeighType.vertex)
    assert neigh.shape == (26, 3)
    assert dist.shape == (26,)
    neigh, dist = calculate_distances_array((1, 1), NeighType.vertex)
    assert neigh.shape == (8, 3)
    assert dist.shape == (8,)
